/* jmixer.cpp - part of MUSE
 * (c)2000 Denis Roio aka jaromil
 * c++ engine for multiple mp3 realtime stream mixing
 */

#include <iostream.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <sys/time.h>
#include <config.h>

#include "jmixer.h"
#include "audioproc.h"

Stream_mixer::Stream_mixer() {

  dsp = 0;
  have_gui = false;

#ifdef MPG123
  
  /* init channels */
  for(c=0;c<MAX_CHANNELS;c++) {
    mp3ch[c].dsp = &dsp;
    mp3ch[c].sample_rate = SAMPLE_RATE;
  }

#endif

  dspout = false;
  icecastout = false;
  fileout = false;
  quit = false;

  /* Initialization as a synchronous object */
  synchronous.registerObjectName("Stream_mixer");

}

void Stream_mixer::run() {

  if(debug)
    cerr << selfName() << " running on thread " << (int)pthread_self() << "\n";

  while(!quit) {
    
    synchronous.select.accept(0, id_stream_mixer);

    synchronous.select.waitUntil(1, LTI::Kernel::now()+5);
    synchronous.wait();

    cafudda();
  }

  cerr << selfName() << " exiting " << "\n";

}

Stream_mixer::~Stream_mixer() {

#ifdef MPG123
  for(i=0;i<MAX_CHANNELS;i++)
    mp3ch[i].stop();
#endif

  if(dsp!=0)
    close_soundcard();
  
  if(icecastout)
    stop_icecast();

}

void Stream_mixer::register_gui(Amuse *reg_gui) { 
  gui = reg_gui; 
  have_gui = true;
  if(HAVE_LAME) {
    char temp[256];
    lame.version();
    sprintf(temp,"Encoding with Lame %s",lame.ver);
    gui->status_lame->value(temp);
  }
}

bool Stream_mixer::open_soundcard()
{
    int format,tstereo,speed,caps;
    static int val;
    
    if((dsp=open("/dev/dsp",O_RDWR))==-1) {
        perror("Error opening /dev/dsp: ");
	return(false);
    } else
      fprintf(stderr,"Found soundcard on /dev/dsp\n");

    /* BUFFER FRAGMENTATION */
    val = (FRAGCOUNT << 16) | FRAGSIZE;

    if(ioctl(dsp,SNDCTL_DSP_SETFRAGMENT,&val)==-1)
      fprintf(stderr,"failed to set dsp buffers\n");

    /* or NE ?*/
    format = AFMT_S16_LE;

    tstereo = 1; /* only stereo _DSP_ in/out 'til anybody complains */
    speed = SAMPLE_RATE; /* 44100hz mixing */

    ioctl(dsp, SNDCTL_DSP_GETCAPS, &caps);
    if(caps & DSP_CAP_DUPLEX) {
      fullduplex = true;
      fprintf(stderr,"full duplex supported. good\n");
      ioctl(dsp, SNDCTL_DSP_SETDUPLEX, 0);
      ioctl(dsp, DSP_CAP_DUPLEX, 0);
    } else {
      fprintf(stderr,"full duplex is not supported\n");
      fullduplex = false;
    }

    /* FORMAT */
    if(ioctl(dsp,SNDCTL_DSP_SETFMT,&format)==-1)
      fprintf(stderr,"failed to set data format\n");

    /* CHANNELS */
    if(ioctl(dsp,SNDCTL_DSP_STEREO,&tstereo)==-1)
      fprintf(stderr,"something went wrong with the stereo setting\n");

    /* SAMPLERATE */
    if(ioctl(dsp,SNDCTL_DSP_SPEED,&speed)==-1)
      fprintf(stderr,"speed setting failed\n");

    /* GET FRAG SIZE BACK */
    if(ioctl(dsp,SNDCTL_DSP_GETBLKSIZE,&val)==-1)
      fprintf(stderr,"get block size failed to return\n");

    fprintf(stderr,"mixing 16bit %dHz stereo\n",speed);

    dspout = false;

    livein = new LiveIn(speed, tstereo+1, &dsp);

    return(true);
} /* open_soundcard */

bool Stream_mixer::set_icecast(char *ip, int port, char *pass, char *mountpoint, int bps, char *name, char *url, char *desc, int login) {

  if(!ice.set_server(ip,port,pass,mountpoint,bps,login)) {
    if(have_gui)
      gui->status_main->value("doom damnit! i can't resolve icecast hostname");
    else
      fprintf(stderr,"Can't resolve icecast server hostname\n");
    synchronous.reply();
    return(false);
  }

  ice.set_meta(name, url, desc);
      
  if((!lame.initialized)||(lame.bps!=bps)) {
    gui->status_main->value("ouch! lame is not initialized!");
    synchronous.reply();
    return(false);
  }
  
  if(ice.start()) { /* we are connected */
    icecastout = true;

    /* notice the user */
    if(have_gui) {
      char temp[256];
      gui->status_main->value("WE GOT MILK !!");
      sprintf(temp,"STREAMING ON [ http://%s:%u/%s ]",ip,port,mountpoint);
      gui->status_ice->value(temp);
    } else
      fprintf(stderr,"Estabilished %ubps stream on %s:%u/%s\n",bps,ip,port,mountpoint);

  } else { /* connection failed */
    
    icecastout = false;
    if(have_gui) {
      char temp[256];
      sprintf(temp,"Got problems streaming on server %s. No milk??",ip);
      gui->status_main->value(temp);
    } else
      fprintf(stderr,"Got problems streaming on server %s. No milk??\n",ip);
  }
  
  synchronous.reply();
  return(icecastout);
}

void Stream_mixer::set_lame(int bps, int freq, int mode, int quality, int lowpass, int highpass) {
  lame.init(bps, freq, mode, quality, lowpass, highpass);
  synchronous.reply();
}

void Stream_mixer::stop_icecast() {
  ice.stop();
  icecastout = false;

  if (have_gui) {
    gui->status_main->value("Stopped streaming to icecast server");
    gui->status_ice->value("");
  } else
    fprintf(stderr,"Stopped streaming to icecast server\n");

  synchronous.reply();
}

int Stream_mixer::set_mp3file(char *file, int bps) {
  int res;
  
  res = mp3file.openmp3(file);
  if(res>0) fileout = true;

  /*  
      if((!lame.initialized)||(lame.bps!=bps))
      lame.init(bps);
  */
  
  synchronous.reply();
  return(res);
}

void Stream_mixer::stop_mp3file() {
  mp3file.closemp3();
  fileout = false;
  synchronous.reply();
}

void Stream_mixer::cafudda()
{
  int i,j;
  int num = 0;
  float res;
  int max = 0;

  /* j is the number of 16bit samples (not counting stereo inrterleaving)*/
  j=BUF_SIZE>>1;
  
  memset(process_buffer,0,BUF_SIZE<<2); /* assuming sizeof(int)==4 */

  have_sound = false;

#ifdef MPG123
  
  for(i=0;i<MAX_CHANNELS;i++) {
    if((mp3ch[i].pid)&&(mp3ch[i].on)) {
      mp3ch[i].get_audio();
      if(mp3ch[i].volume!=1.0)
	for(j=0;j<num;j++)
	  process_buffer[j] = (int) (process_buffer[j] + (mp3ch[i].gotin[j] * mp3ch[i].volume));
      else
	for(j=0;j<num;j--)
	  process_buffer[j] = (int) (process_buffer[j] + mp3ch[i].gotin[j]);

      /* if the mp3 reached the end */
      if(mp3ch[i].end)
	mp3ch[i].ends();
      
      have_sound = true;
    } /* if (mp3ch[i].pid&&mp3ch[i].on) */
  } /* for i=0 to MAX_CHANNELS */

#else

  for(i=0;i<MAX_CHANNELS;i++) {
    /*    if(xach[i].on) {
	  res = xach[i].get_audio();
	  updchan(i,res);      
	  }
    */

    if(xach[i].on) {
      if(xach[i].buffer->nb_channels == 1) {/* mono */
	for(num=0;num<j;num++) {
	  process_buffer[num] = (int) (process_buffer[num] + (((xach[i].buffer->pcm_samples[num])<<8)&(xach[i].buffer->pcm_samples[num]))*xach[i].volume);
	}
      } else {/* stereo */
	res = xach[i].mix(process_buffer);
	updchan(i,res);
	if(max < xach[i].in_smp) max = xach[i].in_smp;
      }

      /* notice that we have pcm to process (at least 1 chan) */
      have_sound = true;
      
    } /* if(xach[i].on) */
  } /* for(i=0;i<MAX_CHANNELS;i++) */
#endif
  
  if(livein->on) {
    livein->get_audio();
    for(num=0;num<j;num++)
      process_buffer[num] = (int) (process_buffer[num] + ((int)livein->gotin[num]));
    /* notice that we have pcm to process (at least 1 chan) */
    have_sound = true;
  }  
  
  if(have_sound) {
    
    /* CLIPPING */
    //    clip_audio(process_buffer,audio_buffer,BUF_SIZE);
    clip_audio(process_buffer,audio_buffer,max);

    /* compute levels */
    if(have_gui)
      //      gui->level = get_level(audio_buffer,BUF_SIZE>>1);
      gui->level = get_level(audio_buffer,max>>1);
    
    /* ENCODE AUDIO */
    if(fileout||icecastout)
      //      lame.encode(&audio_buffer[0],j>>1);
      lame.encode(&audio_buffer[0],max>>2);
    
    /* WRITE 2 MP3 FILE */
    if(fileout)
      mp3file.writemp3();

    /* SEND 2 ICECAST */
    if(icecastout) {
      if(ice.send()!=0) {
	if(have_gui)
	  gui->status_main->value("Send error: cannot send stream to icecast server");
	else
	  fprintf(stderr,"Send error: cannot send stream to icecast server\n");
      }
      
      else if(num==-1) {
	if(have_gui) {
	  fl_alert("Encoding failed: mp3buffer overflowed\nUACK!! bailing out!");
	  exit(-1);
	} else {
	  fprintf(stderr,"Encoding failed: mp3buffer overflowed\nUACK!! bailing out!\n");
	  exit(-1);
	}
      } 
    }
    
    /* WRITE 2 DSP */
    if(dspout)
      //      write(dsp,audio_buffer,BUF_SIZE);
      write(dsp,audio_buffer,max);
    
  } /* if(have_sound) */
  
  /* draw levels */
  if(have_gui)
    gui->mix_monitor->draw_levels(gui->level);
  
}

void Stream_mixer::pause_channel(int ch) {

#ifdef MPG123
  mp3ch[ch].on = !mp3ch[ch].on;
#else
  if(xach[ch].opened)
    xach[ch].on = !xach[ch].on;
#endif

  synchronous.reply();
}

bool Stream_mixer::set_channel(int ch, int playlist_pos) {

  bool res;

#ifdef MPG123
  mp3ch[ch].start(playlist_pos);
#else

  res = xach[ch].set(playlist_pos);

  if(!res) {
    if(have_gui) {
      gui->status_main->value("Doom damnit! can't open the file selected");
    } else {
      fprintf(stderr,"error opening file %s\n", xach[ch].playlist[playlist_pos]);
    }
  } else {
    if(have_gui) {
      char temp[30];
      sprintf(temp,"File loaded on channel [%u]",ch);
      gui->status_main->value(temp);
    }
  }
#endif
  
  synchronous.reply();
  return(res);
}

void Stream_mixer::set_all_volumes(float *vol) {
  int c;
  for(c=0;c<MAX_CHANNELS;c++) {
#ifdef MPG123
    mp3ch[c].volume = vol[c];
#else
    xach[c].volume = vol[c];
#endif
  }
    synchronous.reply();
}

void Stream_mixer::set_volume(int ch, float vol) {

#ifdef MPG123
  mp3ch[ch].volume = vol;
#else
  xach[ch].volume = vol;
#endif

  synchronous.reply();
}

void Stream_mixer::set_speed(int ch, int speed) {

  //#ifdef MPG123
  //  mp3ch[ch].speed = speed;
  //#else
  xach[ch].speed = speed;

  synchronous.reply();
}

bool Stream_mixer::set_playlist(int ch, char *pl) {

  FILE *plist;
  int c = 0;
  char *p;

#ifdef MPG123
  mp3ch[ch].get_playlist(pl);
#else

  xach[ch].get_playlist(pl);

  plist = fopen(pl,"r");
  if(plist == NULL) {
    fprintf(stderr,"I can't read the playlist file!\n");
    return(false);
  }

  while (fgets(&xach[ch].playlist[c][0],MAX_PATH_SIZE,plist) != NULL) {
    p = &xach[ch].playlist[c][0];
    while(*p!='\n') p++;
    *p = '\0';
    if(have_gui) {
      while(*p!='/') p--; /* gives to the gui the filename without path */
      gui->chan[ch]->playlist_browser->add(p+1,NULL);
    }
    c++;
  }

  fclose(plist);
  
  xach[ch].playlist_len = c;
  
#endif

  synchronous.reply();
  return(true);
}

void Stream_mixer::play_channel(int ch) {

#ifdef MPG123
  mp3ch[ch].on = true;
#else
  if(xach[ch].opened)
    xach[ch].play();
#endif

  synchronous.reply();
}
  
bool Stream_mixer::stop_channel(int ch) {

  bool res = true;

  if(ch<=MAX_CHANNELS) {

#ifdef MPG123
    mp3ch[ch].stop();
#else
    res = xach[ch].stop();
#endif
  }

  synchronous.reply();
  return(res);

}

#ifndef MPG123
bool Stream_mixer::set_position(float pos, int ch) {
  bool res = true;

  res = xach[ch].pos(pos);

  synchronous.reply();
  return(res);
}
#endif

void Stream_mixer::dump_status() {

  fprintf(stderr,"__________dumping_status__________\n");

#ifdef MPG123
  for(c=0;c<MAX_CHANNELS;c++) {
    
    if(!mp3ch[c].on)
      fprintf(stderr,"[channel %02u OFF] : PID %u\n",c,mp3ch[c].pid);
    else {
      fprintf(stderr,"[channel %02u ON ] : PID %u\n",c,mp3ch[c].pid);
      fprintf(stderr,"\tname:\t%s\n",mp3ch[c].name);
      fprintf(stderr,"\tvolume:\t%f\n",mp3ch[c].volume);
      fprintf(stderr,"\tspeed:\t%f\n",mp3ch[c].speed);
      fprintf(stderr,"\trate:\t%u\n",mp3ch[c].rate);
      fprintf(stderr,"\tchannels:\t%u\n",mp3ch[c].channels);
      fprintf(stderr,"\tend:\t%u\n",mp3ch[c].end);
      fprintf(stderr,"\tP stream:\t%u\n",mp3ch[c].stream[0]);
      fprintf(stderr,"\tC stream:\t%u\n",mp3ch[c].stream[1]);
    }
  }
#endif

  fprintf(stderr,"__________________________________\n");    
  synchronous.reply();
}

bool Stream_mixer::set_live(bool stat) {
  if(!((dspout)&&(!fullduplex)&&(stat)))
    livein->on = stat;

  synchronous.reply();
  return(livein->on&stat);
}

bool Stream_mixer::set_lineout(bool stat) {
  if(!((livein->on)&&(!fullduplex)&&(stat)))
    dspout = stat;

  synchronous.reply();
  return(dspout&stat);
}

void Stream_mixer::close_soundcard() {
  ioctl(dsp, SNDCTL_DSP_RESET, 0);
  close(dsp);
}

void Stream_mixer::upd_timeline(int ch, float res) {
  char temp[10];
  bool changed = false;
  
  if((res - xach[ch].position)>0.01) {
    xach[ch].position = res;
    changed = true;
  }
  
  if(changed)
    gui->chan[ch]->timeline_slider->value(res);
      
  sprintf(temp,"%02u:%02u:%02u",xach[ch].status->timecode.h,xach[ch].status->timecode.m,xach[ch].status->timecode.s);
  gui->chan[ch]->time->value(temp);

}

void Stream_mixer::updchan(int ch, float res) {

  if(res==3.0) {
    gui->status_main->value("urg... syncing channel");
  } else if(res==2.0) {
    switch(xach[ch].playmode) {
    case 0: /* LOOP */
      gui->chan[ch]->timeline_slider->value(0.0);
      gui->chan[ch]->time->value("00:00:00");
      xach[ch].stop();
      xach[ch].play();
      break;
    case 1: /* CONT */
      int newsong;
      newsong = xach[ch].selected + 1;
      gui->chan[ch]->timeline_slider->value(0.0);
      gui->chan[ch]->time->value("00:00:00");
      gui->chan[ch]->name->value(gui->chan[ch]->playlist_browser->text(newsong+1));
      gui->chan[ch]->playlist_browser->value(newsong+1);
      xach[ch].set(newsong);
      xach[ch].play();
      break;
    case 2: /* PLAY */
      gui->chan[ch]->timeline_slider->value(0.0);
      gui->chan[ch]->time->value("00:00:00");
      xach[ch].stop();
      break;
    }
  } else {
    if(have_gui)
      upd_timeline(ch, res);
  }
}
